from functools import partial
from typing import (Any,
                    Dict)

from aiohttp import ClientSession

from asynctmdb import requests
from asynctmdb.config import API_BASE_URL
from asynctmdb.utils import urljoin

external_sources = {'imdb_id',
                    'freebase_mid', 'freebase_id',
                    'tvdb_id', 'tvrage_id'}


async def by_id(external_id: str,
                *,
                api_base_url: str = API_BASE_URL,
                api_key: str,
                language: str = None,
                external_source: str,
                session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(api_base_url, 'find', external_id)

    params = {}
    if language is not None:
        params['language'] = language

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  external_source=external_source,
                                  **params)
    return response


by = {external_source: partial(by_id,
                               external_source=external_source)
      for external_source in external_sources}



from datetime import datetime

import pytest
from aiohttp import ClientSession

from asynctmdb.common import StatusCode
from asynctmdb.methods.authentication import (create_request_token,
                                              create_session,
                                              create_guest_session,
                                              validate_request_token)
from tests.utils import is_non_empty_hex_string


@pytest.mark.asyncio
async def test_create_request_token(api_base_url: str,
                                    api_key: str,
                                    invalid_api_key: str,
                                    current_date_time: datetime,
                                    session: ClientSession) -> None:
    valid_response = await create_request_token(
            api_base_url=api_base_url,
            api_key=api_key,
            session=session)
    invalid_api_key_response = await create_request_token(
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            session=session)

    succeed = valid_response['success']
    expiration_date_time = valid_response['expires_at']
    request_token = valid_response['request_token']
    invalid_api_key_status_code = invalid_api_key_response['status_code']

    assert isinstance(succeed, bool) and succeed
    assert expiration_date_time >= current_date_time
    assert is_non_empty_hex_string(request_token)
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY


@pytest.mark.asyncio
async def test_validate_request_token(api_base_url: str,
                                      api_key: str,
                                      invalid_api_key: str,
                                      user_name: str,
                                      user_password: str,
                                      unvalidated_request_token: str,
                                      session: ClientSession) -> None:
    valid_response = await validate_request_token(
            api_base_url=api_base_url,
            api_key=api_key,
            username=user_name,
            password=user_password,
            request_token=unvalidated_request_token,
            session=session)
    invalid_api_key_response = await validate_request_token(
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            username=user_name,
            password=user_password,
            request_token=unvalidated_request_token,
            session=session)

    succeed = valid_response['success']
    validated_request_token = valid_response['request_token']
    invalid_api_key_status_code = invalid_api_key_response['status_code']

    assert isinstance(succeed, bool) and succeed
    assert is_non_empty_hex_string(validated_request_token)
    assert validated_request_token == unvalidated_request_token
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY


@pytest.mark.asyncio
async def test_create_session(api_base_url: str,
                              api_key: str,
                              invalid_api_key: str,
                              user_name: str,
                              user_password: str,
                              unvalidated_request_token: str,
                              session: ClientSession) -> None:
    unvalidated_token_response = await create_session(
            api_base_url=api_base_url,
            api_key=api_key,
            request_token=unvalidated_request_token,
            session=session)
    response = await validate_request_token(
            api_base_url=api_base_url,
            api_key=api_key,
            username=user_name,
            password=user_password,
            request_token=unvalidated_request_token,
            session=session)
    validated_request_token = response['request_token']
    valid_response = await create_session(
            api_base_url=api_base_url,
            api_key=api_key,
            request_token=validated_request_token,
            session=session)
    invalid_api_key_response = await create_session(
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            request_token=validated_request_token,
            session=session)

    succeed = valid_response['success']
    session_id = valid_response['session_id']
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    unvalidated_token_status_code = unvalidated_token_response['status_code']

    assert isinstance(succeed, bool) and succeed
    assert is_non_empty_hex_string(session_id)
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert unvalidated_token_status_code == StatusCode.SESSION_DENIED


@pytest.mark.asyncio
async def test_create_guest_session(api_base_url: str,
                                    api_key: str,
                                    invalid_api_key: str,
                                    session: ClientSession) -> None:
    valid_response = await create_guest_session(
            api_base_url=api_base_url,
            api_key=api_key,
            session=session)
    invalid_api_key_response = await create_guest_session(
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            session=session)

    succeed = valid_response['success']
    expiration_date_time = valid_response['expires_at']
    guest_session_id = valid_response['guest_session_id']
    invalid_api_key_status_code = invalid_api_key_response['status_code']

    assert succeed
    assert expiration_date_time >= datetime.utcnow()
    assert isinstance(guest_session_id, str)
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY



from .api import (api_base_urls,
                  api_keys,
                  invalid_api_keys)
from .authentication import (users_names,
                             users_passwords)
from .imdb import invalid_imdb_ids
from .movies import (MIN_MOVIE_ID,
                     movies_ids,
                     nonexistent_movies_ids,
                     movies_changes_date_times,
                     movies_ratings,
                     pages_numbers,
                     invalid_pages_numbers)
from .utils import TMDB_FOUNDATION_DATE



import operator
from datetime import (date,
                      datetime)
from itertools import chain
from typing import Set

import pytest
from aiohttp import ClientSession

from asynctmdb.common import StatusCode
from asynctmdb.methods import movies
from tests.utils import (is_positive_integer,
                         is_non_empty_string,
                         is_valid_paginated_record,
                         are_valid_results)


@pytest.mark.asyncio
async def test_movie_details(movie_id: int,
                             nonexistent_movie_id: int,
                             min_movie_id: int,
                             imdb_id: str,
                             api_base_url: str,
                             api_key: str,
                             invalid_api_key: str,
                             session: ClientSession) -> None:
    record = await movies.details(
            movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            session=session)
    invalid_api_key_response = await movies.details(
            movie_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            session=session)
    nonexistent_movie_response = await movies.details(
            nonexistent_movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            session=session)

    record_id = record['id']
    record_imdb_id = record['imdb_id']
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    nonexistent_movie_status_code = nonexistent_movie_response['status_code']

    assert isinstance(record, dict)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(nonexistent_movie_response, dict)
    assert record_imdb_id == imdb_id
    assert is_positive_integer(record_id)
    assert record_id >= min_movie_id
    assert record_id == movie_id
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert nonexistent_movie_status_code == StatusCode.RESOURCE_NOT_FOUND


@pytest.mark.asyncio
async def test_account_movie_states(movie_id: int,
                                    min_movie_id: int,
                                    api_base_url: str,
                                    api_key: str,
                                    invalid_api_key: str,
                                    session_id: str,
                                    invalid_session_id: str,
                                    session: ClientSession) -> None:
    record = await movies.account_states(
            movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            session_id=session_id,
            session=session)
    invalid_api_key_response = await movies.account_states(
            movie_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            session_id=session_id,
            session=session)
    invalid_session_response = await movies.account_states(
            movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            session_id=invalid_session_id,
            session=session)

    record_id = record['id']
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    invalid_session_status_code = invalid_session_response['status_code']

    assert isinstance(record, dict)
    assert isinstance(invalid_api_key_response, dict)
    assert is_positive_integer(record_id)
    assert record_id >= min_movie_id
    assert record_id == movie_id
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert invalid_session_status_code == StatusCode.AUTHENTICATION_FAILED


@pytest.mark.asyncio
async def test_movie_alternative_titles(movie_id: int,
                                        nonexistent_movie_id: int,
                                        min_movie_id: int,
                                        api_base_url: str,
                                        api_key: str,
                                        invalid_api_key: str,
                                        session: ClientSession) -> None:
    record = await movies.alternative_titles(
            movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            session=session)
    invalid_api_key_response = await movies.alternative_titles(
            movie_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            session=session)
    nonexistent_movie_response = await movies.alternative_titles(
            nonexistent_movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            session=session)

    record_id = record['id']
    record_titles = map(operator.itemgetter('title'), record['titles'])
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    nonexistent_movie_status_code = nonexistent_movie_response['status_code']

    assert isinstance(record, dict)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(nonexistent_movie_response, dict)
    assert is_positive_integer(record_id)
    assert record_id >= min_movie_id
    assert record_id == movie_id
    assert all(map(is_non_empty_string, record_titles))
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert nonexistent_movie_status_code == StatusCode.RESOURCE_NOT_FOUND


@pytest.mark.asyncio
async def test_movie_changes(movie_id: int,
                             nonexistent_movie_id: int,
                             movie_changes_start_date: datetime,
                             movie_changes_end_date: datetime,
                             page_number: int,
                             invalid_page_number: int,
                             api_base_url: str,
                             api_key: str,
                             invalid_api_key: str,
                             tmdb_foundation_date: date,
                             session: ClientSession) -> None:
    empty_interval_record = await movies.changes(
            movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            start_date=datetime.max,
            end_date=datetime.min,
            page=page_number,
            session=session)
    nonexistent_movie_record = await movies.changes(
            nonexistent_movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            start_date=movie_changes_start_date,
            end_date=movie_changes_end_date,
            page=page_number,
            session=session)
    record = await movies.changes(
            movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            start_date=movie_changes_start_date,
            end_date=movie_changes_end_date,
            page=page_number,
            session=session)
    invalid_api_key_response = await movies.changes(
            movie_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            start_date=movie_changes_start_date,
            end_date=movie_changes_end_date,
            page=page_number,
            session=session)
    invalid_page_response = await movies.changes(
            movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            start_date=movie_changes_start_date,
            end_date=movie_changes_end_date,
            page=invalid_page_number,
            session=session)

    empty_interval_results = empty_interval_record['changes']
    nonexistent_movie_results = nonexistent_movie_record['changes']
    results = record['changes']
    items = chain.from_iterable(map(operator.itemgetter('items'),
                                    results))
    changes_times = map(operator.itemgetter('time'), items)
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    invalid_page_status_code = invalid_page_response['status_code']

    assert isinstance(empty_interval_record, dict)
    assert isinstance(nonexistent_movie_record, dict)
    assert isinstance(record, dict)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(invalid_page_response, dict)
    assert (isinstance(empty_interval_results, list) and
            not empty_interval_results)
    assert (isinstance(nonexistent_movie_results, list) and
            not nonexistent_movie_results)
    assert are_valid_results(results)
    assert all(isinstance(change_time, datetime) and
               change_time.date() > tmdb_foundation_date
               for change_time in changes_times)
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert invalid_page_status_code == StatusCode.INVALID_PAGE


@pytest.mark.asyncio
async def test_movie_credits(movie_id: int,
                             nonexistent_movie_id: int,
                             min_movie_id: int,
                             api_base_url: str,
                             api_key: str,
                             invalid_api_key: str,
                             session: ClientSession) -> None:
    record = await movies.credits(movie_id,
                                  api_base_url=api_base_url,
                                  api_key=api_key,
                                  session=session)
    invalid_api_key_response = await movies.credits(
            movie_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            session=session)
    nonexistent_movie_response = await movies.credits(
            nonexistent_movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            session=session)

    record_id = record['id']
    staff_records = record['cast'] + record['crew']
    staff_records_ids = map(operator.itemgetter('id'), staff_records)
    staff_records_names = map(operator.itemgetter('name'), staff_records)
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    nonexistent_movie_status_code = nonexistent_movie_response['status_code']

    assert isinstance(record, dict)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(nonexistent_movie_response, dict)
    assert is_positive_integer(record_id)
    assert record_id >= min_movie_id
    assert record_id == movie_id
    assert isinstance(staff_records, list)
    assert all(map(is_positive_integer, staff_records_ids))
    assert all(map(is_non_empty_string, staff_records_names))
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert nonexistent_movie_status_code == StatusCode.RESOURCE_NOT_FOUND


@pytest.mark.asyncio
async def test_movie_images(movie_id: int,
                            nonexistent_movie_id: int,
                            min_movie_id: int,
                            api_base_url: str,
                            api_key: str,
                            invalid_api_key: str,
                            session: ClientSession) -> None:
    record = await movies.images(
            movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            session=session)
    invalid_api_key_response = await movies.images(
            movie_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            session=session)
    nonexistent_movie_response = await movies.images(
            nonexistent_movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            session=session)

    record_id = record['id']
    results = record['backdrops']
    backdrops_records_heights = map(operator.itemgetter('height'),
                                    results)
    backdrops_records_widths = map(operator.itemgetter('width'),
                                   results)
    backdrops_records_files_paths = map(operator.itemgetter('file_path'),
                                        results)
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    nonexistent_movie_status_code = nonexistent_movie_response['status_code']

    assert isinstance(record, dict)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(nonexistent_movie_response, dict)
    assert is_positive_integer(record_id)
    assert record_id >= min_movie_id
    assert record_id == movie_id
    assert are_valid_results(results)
    assert all(map(is_positive_integer, backdrops_records_heights))
    assert all(map(is_positive_integer, backdrops_records_widths))
    assert all(map(is_non_empty_string, backdrops_records_files_paths))
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert nonexistent_movie_status_code == StatusCode.RESOURCE_NOT_FOUND


@pytest.mark.asyncio
async def test_movie_keywords(movie_id: int,
                              nonexistent_movie_id: int,
                              min_movie_id: int,
                              api_base_url: str,
                              api_key: str,
                              invalid_api_key: str,
                              session: ClientSession) -> None:
    record = await movies.keywords(movie_id,
                                   api_base_url=api_base_url,
                                   api_key=api_key,
                                   session=session)
    invalid_api_key_response = await movies.keywords(
            movie_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            session=session)
    nonexistent_movie_response = await movies.keywords(
            nonexistent_movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            session=session)

    record_id = record['id']
    results = record['keywords']
    keywords_records_ids = map(operator.itemgetter('id'),
                               results)
    keywords_records_names = map(operator.itemgetter('name'),
                                 results)
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    nonexistent_movie_status_code = nonexistent_movie_response['status_code']

    assert isinstance(record, dict)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(nonexistent_movie_response, dict)
    assert is_positive_integer(record_id)
    assert record_id >= min_movie_id
    assert record_id == movie_id
    assert are_valid_results(results)
    assert all(map(is_positive_integer, keywords_records_ids))
    assert all(map(is_non_empty_string, keywords_records_names))
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert nonexistent_movie_status_code == StatusCode.RESOURCE_NOT_FOUND


@pytest.mark.asyncio
async def test_movie_release_dates(movie_id: int,
                                   nonexistent_movie_id: int,
                                   min_movie_id: int,
                                   min_movie_release_date: date,
                                   api_base_url: str,
                                   api_key: str,
                                   invalid_api_key: str,
                                   session: ClientSession) -> None:
    record = await movies.release_dates(movie_id,
                                        api_base_url=api_base_url,
                                        api_key=api_key,
                                        session=session)
    invalid_api_key_response = await movies.release_dates(
            movie_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            session=session)
    nonexistent_movie_response = await movies.release_dates(
            nonexistent_movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            session=session)

    record_id = record['id']
    results = record['results']
    release_dates_records = chain.from_iterable(
            map(movies.release_dates_record, results))
    release_dates = map(operator.itemgetter('release_date'),
                        release_dates_records)
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    nonexistent_movie_status_code = nonexistent_movie_response['status_code']

    assert isinstance(record, dict)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(nonexistent_movie_response, dict)
    assert is_positive_integer(record_id)
    assert record_id >= min_movie_id
    assert record_id == movie_id
    assert are_valid_results(results)
    assert all(min_movie_release_date <= release_date.date()
               for release_date in release_dates)
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert nonexistent_movie_status_code == StatusCode.RESOURCE_NOT_FOUND


@pytest.mark.asyncio
async def test_movie_videos(movie_id: int,
                            nonexistent_movie_id: int,
                            min_movie_id: int,
                            movie_videos_sites: Set[str],
                            api_base_url: str,
                            api_key: str,
                            invalid_api_key: str,
                            session: ClientSession) -> None:
    record = await movies.videos(movie_id,
                                 api_base_url=api_base_url,
                                 api_key=api_key,
                                 session=session)
    invalid_api_key_response = await movies.videos(
            movie_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            session=session)
    nonexistent_movie_response = await movies.videos(
            nonexistent_movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            session=session)

    record_id = record['id']
    results = record['results']
    sites = map(operator.itemgetter('site'), results)
    sizes = map(operator.itemgetter('size'), results)
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    nonexistent_movie_status_code = nonexistent_movie_response['status_code']

    assert isinstance(record, dict)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(nonexistent_movie_response, dict)
    assert is_positive_integer(record_id)
    assert record_id >= min_movie_id
    assert record_id == movie_id
    assert are_valid_results(results)
    assert all(site in movie_videos_sites
               for site in sites)
    assert all(map(is_positive_integer, sizes))
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert nonexistent_movie_status_code == StatusCode.RESOURCE_NOT_FOUND


@pytest.mark.asyncio
async def test_movie_translations(movie_id: int,
                                  nonexistent_movie_id: int,
                                  min_movie_id: int,
                                  api_base_url: str,
                                  api_key: str,
                                  invalid_api_key: str,
                                  session: ClientSession) -> None:
    record = await movies.translations(movie_id,
                                       api_base_url=api_base_url,
                                       api_key=api_key,
                                       session=session)
    invalid_api_key_response = await movies.translations(
            movie_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            session=session)
    nonexistent_movie_response = await movies.translations(
            nonexistent_movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            session=session)

    record_id = record['id']
    results = record['translations']
    names = map(operator.itemgetter('name'), results)
    english_names = map(operator.itemgetter('english_name'),
                        results)
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    nonexistent_movie_status_code = nonexistent_movie_response['status_code']

    assert isinstance(record, dict)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(nonexistent_movie_response, dict)
    assert is_positive_integer(record_id)
    assert record_id >= min_movie_id
    assert record_id == movie_id
    assert are_valid_results(results)
    assert all(map(is_non_empty_string, names))
    assert all(map(is_non_empty_string, english_names))
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert nonexistent_movie_status_code == StatusCode.RESOURCE_NOT_FOUND


@pytest.mark.asyncio
async def test_movie_recommendations(movie_id: int,
                                     nonexistent_movie_id: int,
                                     page_number: int,
                                     invalid_page_number: int,
                                     min_movie_release_date: date,
                                     api_base_url: str,
                                     api_key: str,
                                     invalid_api_key: str,
                                     session: ClientSession) -> None:
    record = await movies.recommendations(movie_id,
                                          api_base_url=api_base_url,
                                          api_key=api_key,
                                          page=page_number,
                                          session=session)
    invalid_api_key_response = await movies.recommendations(
            movie_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            page=page_number,
            session=session)
    nonexistent_movie_response = await movies.recommendations(
            nonexistent_movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            page=page_number,
            session=session)
    invalid_page_response = await movies.recommendations(
            movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            page=invalid_page_number,
            session=session)

    results = record['results']
    release_dates = map(operator.itemgetter('release_date'), results)
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    nonexistent_movie_status_code = nonexistent_movie_response['status_code']
    invalid_page_status_code = invalid_page_response['status_code']

    assert is_valid_paginated_record(record)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(nonexistent_movie_response, dict)
    assert isinstance(invalid_page_response, dict)
    assert record['page'] == page_number
    assert are_valid_results(results)
    assert all(release_date is None or
               isinstance(release_date, date) and
               release_date > min_movie_release_date
               for release_date in release_dates)
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert nonexistent_movie_status_code == StatusCode.INTERNAL_ERROR
    assert invalid_page_status_code == StatusCode.INVALID_PAGE


@pytest.mark.asyncio
async def test_similar_movies(movie_id: int,
                              nonexistent_movie_id: int,
                              page_number: int,
                              invalid_page_number: int,
                              min_movie_release_date: date,
                              api_base_url: str,
                              api_key: str,
                              invalid_api_key: str,
                              session: ClientSession) -> None:
    record = await movies.similar(movie_id,
                                  api_base_url=api_base_url,
                                  api_key=api_key,
                                  page=page_number,
                                  session=session)
    invalid_api_key_response = await movies.similar(
            movie_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            page=page_number,
            session=session)
    nonexistent_movie_response = await movies.similar(
            nonexistent_movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            page=page_number,
            session=session)
    invalid_page_response = await movies.similar(
            movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            page=invalid_page_number,
            session=session)

    results = record['results']
    release_dates = map(operator.itemgetter('release_date'), results)
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    nonexistent_movie_status_code = nonexistent_movie_response['status_code']
    invalid_page_status_code = invalid_page_response['status_code']

    assert is_valid_paginated_record(record)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(nonexistent_movie_response, dict)
    assert isinstance(invalid_page_response, dict)
    assert record['page'] == page_number
    assert are_valid_results(results)
    assert all(release_date is None or
               isinstance(release_date, date) and
               release_date > min_movie_release_date
               for release_date in release_dates)
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert nonexistent_movie_status_code == StatusCode.RESOURCE_NOT_FOUND
    assert invalid_page_status_code == StatusCode.INVALID_PAGE


@pytest.mark.asyncio
async def test_movie_reviews(movie_id: int,
                             nonexistent_movie_id: int,
                             min_movie_id: int,
                             page_number: int,
                             invalid_page_number: int,
                             api_base_url: str,
                             api_key: str,
                             invalid_api_key: str,
                             session: ClientSession) -> None:
    record = await movies.reviews(movie_id,
                                  api_base_url=api_base_url,
                                  api_key=api_key,
                                  page=page_number,
                                  session=session)
    invalid_api_key_response = await movies.reviews(
            movie_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            page=page_number,
            session=session)
    nonexistent_movie_response = await movies.reviews(
            nonexistent_movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            page=page_number,
            session=session)
    invalid_page_response = await movies.reviews(
            movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            page=invalid_page_number,
            session=session)

    record_id = record['id']
    results = record['results']
    urls = map(operator.itemgetter('url'), results)
    reviews_ids = map(operator.itemgetter('id'), results)
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    nonexistent_movie_status_code = nonexistent_movie_response['status_code']
    invalid_page_status_code = invalid_page_response['status_code']

    assert is_valid_paginated_record(record)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(nonexistent_movie_response, dict)
    assert isinstance(invalid_page_response, dict)
    assert record['page'] == page_number
    assert is_positive_integer(record_id)
    assert record_id >= min_movie_id
    assert record_id == movie_id
    assert are_valid_results(results)
    assert all(url == f'https://www.themoviedb.org/review/{review_id}'
               for review_id, url in zip(reviews_ids, urls))
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert nonexistent_movie_status_code == StatusCode.RESOURCE_NOT_FOUND
    assert invalid_page_status_code == StatusCode.INVALID_PAGE


@pytest.mark.asyncio
async def test_movie_lists(movie_id: int,
                           nonexistent_movie_id: int,
                           min_movie_id: int,
                           page_number: int,
                           invalid_page_number: int,
                           api_base_url: str,
                           api_key: str,
                           invalid_api_key: str,
                           session: ClientSession) -> None:
    record = await movies.lists(movie_id,
                                api_base_url=api_base_url,
                                api_key=api_key,
                                page=page_number,
                                session=session)
    invalid_api_key_response = await movies.lists(
            movie_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            page=page_number,
            session=session)
    nonexistent_movie_response = await movies.lists(
            nonexistent_movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            page=page_number,
            session=session)
    invalid_page_response = await movies.lists(
            movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            page=invalid_page_number,
            session=session)

    record_id = record['id']
    results = record['results']
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    nonexistent_movie_status_code = nonexistent_movie_response['status_code']
    invalid_page_status_code = invalid_page_response['status_code']

    assert is_valid_paginated_record(record)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(nonexistent_movie_response, dict)
    assert isinstance(invalid_page_response, dict)
    assert record['page'] == page_number
    assert is_positive_integer(record_id)
    assert record_id >= min_movie_id
    assert record_id == movie_id
    assert are_valid_results(results)
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert nonexistent_movie_status_code == StatusCode.INTERNAL_ERROR
    assert invalid_page_status_code == StatusCode.INVALID_PAGE


@pytest.mark.asyncio
async def test_rate_movie(movie_id: int,
                          nonexistent_movie_id: int,
                          movie_rating: float,
                          guest_session_id: str,
                          invalid_session_id: str,
                          api_base_url: str,
                          api_key: str,
                          invalid_api_key: str,
                          session: ClientSession) -> None:
    valid_response = await movies.rate(
            movie_id,
            rating=movie_rating,
            api_base_url=api_base_url,
            api_key=api_key,
            guest_session_id=guest_session_id,
            session=session)
    invalid_api_key_response = await movies.rate(
            movie_id,
            rating=movie_rating,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            guest_session_id=guest_session_id,
            session=session)
    invalid_session_response = await movies.rate(
            movie_id,
            rating=movie_rating,
            api_base_url=api_base_url,
            api_key=api_key,
            session_id=invalid_session_id,
            session=session)
    nonexistent_movie_response = await movies.rate(
            nonexistent_movie_id,
            rating=movie_rating,
            api_base_url=api_base_url,
            api_key=api_key,
            guest_session_id=guest_session_id,
            session=session)

    valid_status_code = valid_response['status_code']
    invalid_session_status_code = invalid_session_response['status_code']
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    nonexistent_movie_status_code = nonexistent_movie_response['status_code']

    assert isinstance(valid_response, dict)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(nonexistent_movie_response, dict)
    assert valid_status_code == StatusCode.SUCCESS
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert invalid_session_status_code == StatusCode.AUTHENTICATION_FAILED
    assert nonexistent_movie_status_code == StatusCode.RESOURCE_NOT_FOUND


@pytest.mark.asyncio
async def test_delete_rating(movie_id: int,
                             nonexistent_movie_id: int,
                             guest_session_id: str,
                             invalid_session_id: str,
                             api_base_url: str,
                             api_key: str,
                             invalid_api_key: str,
                             session: ClientSession) -> None:
    valid_response = await movies.delete_rating(
            movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            guest_session_id=guest_session_id,
            session=session)
    invalid_api_key_response = await movies.delete_rating(
            movie_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            guest_session_id=guest_session_id,
            session=session)
    invalid_session_response = await movies.delete_rating(
            movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            session_id=invalid_session_id,
            session=session)
    nonexistent_movie_response = await movies.delete_rating(
            nonexistent_movie_id,
            api_base_url=api_base_url,
            api_key=api_key,
            guest_session_id=guest_session_id,
            session=session)

    valid_status_code = valid_response['status_code']
    invalid_session_status_code = invalid_session_response['status_code']
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    nonexistent_movie_status_code = nonexistent_movie_response['status_code']

    assert isinstance(valid_response, dict)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(nonexistent_movie_response, dict)
    assert valid_status_code == StatusCode.SUCCESSFULLY_DELETED
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert invalid_session_status_code == StatusCode.AUTHENTICATION_FAILED
    assert nonexistent_movie_status_code == StatusCode.RESOURCE_NOT_FOUND


@pytest.mark.asyncio
async def test_latest_movie(min_movie_id: int,
                            api_base_url: str,
                            api_key: str,
                            invalid_api_key: str,
                            session: ClientSession) -> None:
    record = await movies.latest(
            api_base_url=api_base_url,
            api_key=api_key,
            session=session)
    invalid_api_key_response = await movies.latest(
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            session=session)

    record_id = record['id']
    invalid_api_key_status_code = invalid_api_key_response['status_code']

    assert isinstance(record, dict)
    assert isinstance(invalid_api_key_response, dict)
    assert is_positive_integer(record_id)
    assert record_id >= min_movie_id
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY


@pytest.mark.asyncio
async def test_now_playing_movies(page_number: int,
                                  invalid_page_number: int,
                                  api_base_url: str,
                                  api_key: str,
                                  invalid_api_key: str,
                                  session: ClientSession) -> None:
    record = await movies.now_playing(
            api_base_url=api_base_url,
            api_key=api_key,
            page=page_number,
            session=session)
    invalid_api_key_response = await movies.now_playing(
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            page=page_number,
            session=session)
    invalid_page_response = await movies.now_playing(
            api_base_url=api_base_url,
            api_key=api_key,
            page=invalid_page_number,
            session=session)

    dates = record['dates']
    results = record['results']
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    invalid_page_errors = invalid_page_response['errors']

    assert is_valid_paginated_record(record)
    assert isinstance(dates, dict)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(invalid_page_response, dict)
    assert record['page'] == page_number
    assert are_valid_results(results)
    assert all(isinstance(dates[key], date)
               for key in movies.NOW_PLAYING_DATES_KEYS)
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert (isinstance(invalid_page_errors, list) and
            len(invalid_page_errors) == 1)


@pytest.mark.asyncio
async def test_popular_movies(page_number: int,
                              invalid_page_number: int,
                              api_base_url: str,
                              api_key: str,
                              invalid_api_key: str,
                              session: ClientSession) -> None:
    record = await movies.popular(
            api_base_url=api_base_url,
            api_key=api_key,
            page=page_number,
            session=session)
    invalid_api_key_response = await movies.popular(
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            page=page_number,
            session=session)
    invalid_page_response = await movies.popular(
            api_base_url=api_base_url,
            api_key=api_key,
            page=invalid_page_number,
            session=session)

    results = record['results']
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    invalid_page_errors = invalid_page_response['errors']

    assert is_valid_paginated_record(record)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(invalid_page_response, dict)
    assert record['page'] == page_number
    assert are_valid_results(results)
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert (isinstance(invalid_page_errors, list) and
            len(invalid_page_errors) == 1)


@pytest.mark.asyncio
async def test_top_rated_movies(page_number: int,
                                invalid_page_number: int,
                                api_base_url: str,
                                api_key: str,
                                invalid_api_key: str,
                                session: ClientSession) -> None:
    record = await movies.top_rated(
            api_base_url=api_base_url,
            api_key=api_key,
            page=page_number,
            session=session)
    invalid_api_key_response = await movies.top_rated(
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            page=page_number,
            session=session)
    invalid_page_response = await movies.top_rated(
            api_base_url=api_base_url,
            api_key=api_key,
            page=invalid_page_number,
            session=session)

    results = record['results']
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    invalid_page_errors = invalid_page_response['errors']

    assert is_valid_paginated_record(record)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(invalid_page_response, dict)
    assert record['page'] == page_number
    assert are_valid_results(results)
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert (isinstance(invalid_page_errors, list) and
            len(invalid_page_errors) == 1)


@pytest.mark.asyncio
async def test_upcoming_movies(page_number: int,
                               invalid_page_number: int,
                               api_base_url: str,
                               api_key: str,
                               invalid_api_key: str,
                               session: ClientSession) -> None:
    record = await movies.upcoming(
            api_base_url=api_base_url,
            api_key=api_key,
            page=page_number,
            session=session)
    invalid_api_key_response = await movies.upcoming(
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            page=page_number,
            session=session)
    invalid_page_response = await movies.upcoming(
            api_base_url=api_base_url,
            api_key=api_key,
            page=invalid_page_number,
            session=session)

    results = record['results']
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    invalid_page_errors = invalid_page_response['errors']

    assert is_valid_paginated_record(record)
    assert isinstance(invalid_api_key_response, dict)
    assert isinstance(invalid_page_response, dict)
    assert record['page'] == page_number
    assert are_valid_results(results)
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY
    assert (isinstance(invalid_page_errors, list) and
            len(invalid_page_errors) == 1)



from asyncio import (AbstractEventLoop,
                     ensure_future)

import pytest
from aiohttp import ClientSession

from asynctmdb.methods.authentication import (
    create_request_token,
    create_session,
    validate_request_token,
    create_guest_session)
from tests import strategies
from tests.utils import example


@pytest.fixture(scope='function')
def user_name() -> str:
    return example(strategies.users_names)


@pytest.fixture(scope='function')
def user_password() -> str:
    return example(strategies.users_passwords)


@pytest.fixture(scope='function')
def unvalidated_request_token(api_base_url: str,
                              api_key: str,
                              session: ClientSession,
                              event_loop: AbstractEventLoop) -> str:
    future = ensure_future(create_request_token(
            api_base_url=api_base_url,
            api_key=api_key,
            session=session))
    response = event_loop.run_until_complete(future)
    return response['request_token']


@pytest.fixture(scope='function')
def guest_session_id(api_base_url: str,
                     api_key: str,
                     session: ClientSession,
                     event_loop: AbstractEventLoop) -> str:
    future = ensure_future(create_guest_session(
            api_base_url=api_base_url,
            api_key=api_key,
            session=session))
    response = event_loop.run_until_complete(future)
    return response['guest_session_id']


@pytest.fixture(scope='function')
def session_id(api_base_url: str,
               api_key: str,
               user_name: str,
               user_password: str,
               unvalidated_request_token: str,
               session: ClientSession,
               event_loop: AbstractEventLoop) -> str:
    future = ensure_future(validate_request_token(
            api_base_url=api_base_url,
            api_key=api_key,
            username=user_name,
            password=user_password,
            request_token=unvalidated_request_token,
            session=session))
    response = event_loop.run_until_complete(future)
    validated_request_token = response['request_token']
    future = ensure_future(create_session(
            api_base_url=api_base_url,
            api_key=api_key,
            request_token=validated_request_token,
            session=session))
    response = event_loop.run_until_complete(future)
    return response['session_id']


@pytest.fixture(scope='session')
def invalid_session_id() -> str:
    return ''



from asyncio import (get_event_loop,
                     ensure_future)
from contextlib import closing
from datetime import date

from aiohttp import ClientSession

from asynctmdb.methods import movies

TMDB_FOUNDATION_DATE = date(year=2008,
                            month=1,
                            day=1)


def load_max_movie_id(api_base_url: str,
                      api_key: str) -> int:
    loop = get_event_loop()
    session = ClientSession(loop=loop)
    with closing(session):
        future = ensure_future(movies.latest(api_base_url=api_base_url,
                                             api_key=api_key,
                                             session=session))
        result = loop.run_until_complete(future)
    return result['id']



from datetime import date

import pytest
from aiohttp import ClientSession

from asynctmdb import imdb
from asynctmdb.common import StatusCode
from tests.utils import is_positive_integer


@pytest.mark.asyncio
async def test_find_by_imdb_id(imdb_id: str,
                               invalid_imdb_id: str,
                               non_existent_imdb_id: str,
                               min_movie_id: int,
                               min_movie_release_date: date,
                               api_base_url: str,
                               api_key: str,
                               invalid_api_key: str,
                               session: ClientSession) -> None:
    record = await imdb.find.movie(imdb_id,
                                   api_base_url=api_base_url,
                                   api_key=api_key,
                                   session=session)
    invalid_api_key_response = await imdb.find.movie(
            imdb_id,
            api_base_url=api_base_url,
            api_key=invalid_api_key,
            session=session)

    record_id = record['id']
    invalid_api_key_status_code = invalid_api_key_response['status_code']
    release_date = record['release_date']

    assert isinstance(record, dict)
    assert isinstance(invalid_api_key_response, dict)
    assert is_positive_integer(record_id)
    assert record_id >= min_movie_id
    assert (release_date is None or
            isinstance(release_date, date) and
            release_date >= min_movie_release_date)
    assert invalid_api_key_status_code == StatusCode.INVALID_API_KEY

    with pytest.raises(ValueError):
        await imdb.find.movie(invalid_imdb_id,
                              api_base_url=api_base_url,
                              api_key=api_key,
                              session=session)
    with pytest.raises(ValueError):
        await imdb.find.movie(non_existent_imdb_id,
                              api_base_url=api_base_url,
                              api_key=api_key,
                              session=session)



from typing import (Union,
                    Dict,
                    List)

from aiohttp import ClientSession

from asynctmdb import requests
from asynctmdb.config import API_BASE_URL
from asynctmdb.utils import urljoin

GenresType = List[Dict[str, Union[int, str]]]


async def movie(*,
                api_base_url: str = API_BASE_URL,
                api_key: str,
                language: str = None,
                session: ClientSession) -> GenresType:
    method_url = urljoin(base_method_url(api_base_url), 'movie/list')

    params = {}
    if language is not None:
        params['language'] = language

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    return response['genres']


async def tv(*,
             api_base_url: str = API_BASE_URL,
             api_key: str,
             language: str = None,
             session: ClientSession) -> GenresType:
    method_url = urljoin(base_method_url(api_base_url), 'tv/list')

    params = {}
    if language is not None:
        params['language'] = language

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    return response['genres']


def base_method_url(api_base_url: str) -> str:
    return urljoin(api_base_url, 'genre')



from datetime import (date,
                      datetime)

import pytest

from tests import strategies


@pytest.fixture(scope='function')
def current_date_time() -> datetime:
    return datetime.utcnow()


@pytest.fixture(scope='session')
def tmdb_foundation_date() -> date:
    # TODO: find more accurate date
    return strategies.TMDB_FOUNDATION_DATE



from asyncio import AbstractEventLoop
from typing import Generator

import pytest
from aiohttp import ClientSession


@pytest.fixture(scope='function')
def session(event_loop: AbstractEventLoop
            ) -> Generator[ClientSession, None, None]:
    with ClientSession(loop=event_loop) as result:
        yield result



from typing import (Any,
                    Dict)

import pytest

from asynctmdb import imdb
from tests import strategies
from tests.utils import example


@pytest.fixture(scope='function')
def imdb_id(movie_details: Dict[str, Any]) -> str:
    return movie_details['imdb_id']


@pytest.fixture(scope='session')
def non_existent_imdb_id() -> str:
    return imdb.title_id.from_int(0)


@pytest.fixture(scope='function')
def invalid_imdb_id() -> str:
    return example(strategies.invalid_imdb_ids)



from functools import partial
from typing import (Any,
                    Callable,
                    Dict)

from aiohttp import ClientSession

from .types import AsyncContextManager


async def send(*,
               method: Callable[[ClientSession, str, Any],
                                AsyncContextManager],
               method_url: str,
               session: ClientSession,
               json_body: Dict[str, Any] = None,
               **params: Dict[str, str]
               ) -> Dict[str, Any]:
    async with method(session,
                      method_url,
                      json=json_body,
                      params=params) as response:
        response_json = await response.json()
        return response_json


get = partial(send,
              method=ClientSession.get)
post = partial(send,
               method=ClientSession.post)
delete = partial(send,
                 method=ClientSession.delete)



from enum import IntEnum

DATE_FORMAT = '%Y-%m-%d'
TIME_FORMAT = '%H:%M:%S'
DATE_TIME_FORMAT = f'{DATE_FORMAT} {TIME_FORMAT} %Z'


class StatusCode(IntEnum):
    SUCCESS = 1
    AUTHENTICATION_FAILED = 3
    INVALID_API_KEY = 7
    INTERNAL_ERROR = 11
    SUCCESSFULLY_DELETED = 13
    SESSION_DENIED = 17
    INVALID_PAGE = 22
    RESOURCE_NOT_FOUND = 34



from datetime import datetime
from typing import (Any,
                    Union,
                    Dict)

from aiohttp import ClientSession

from asynctmdb import requests
from asynctmdb.common import DATE_TIME_FORMAT
from asynctmdb.config import API_BASE_URL
from asynctmdb.utils import urljoin


async def create_request_token(*,
                               api_base_url: str = API_BASE_URL,
                               api_key: str,
                               session: ClientSession,
                               date_time_format: str = DATE_TIME_FORMAT
                               ) -> Dict[str, Union[int, str, datetime]]:
    method_url = urljoin(base_method_url(api_base_url),
                         'token',
                         'new')
    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key)
    normalize_response(response,
                       date_time_format=date_time_format)
    return response


async def create_session(*,
                         api_base_url: str = API_BASE_URL,
                         api_key: str,
                         request_token: str,
                         session: ClientSession
                         ) -> Dict[str, Union[int, str]]:
    method_url = urljoin(base_method_url(api_base_url),
                         'session',
                         'new')
    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  request_token=request_token)
    return response


async def validate_request_token(*,
                                 api_base_url: str = API_BASE_URL,
                                 api_key: str,
                                 username: str,
                                 password: str,
                                 request_token: str,
                                 session: ClientSession
                                 ) -> Dict[str, Union[int, str]]:
    method_url = urljoin(base_method_url(api_base_url),
                         'token',
                         'validate_with_login')
    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  username=username,
                                  password=password,
                                  request_token=request_token)
    return response


async def create_guest_session(*,
                               api_base_url: str = API_BASE_URL,
                               api_key: str,
                               session: ClientSession,
                               date_time_format: str = DATE_TIME_FORMAT
                               ) -> Dict[str, Union[int, str, datetime]]:
    method_url = urljoin(base_method_url(api_base_url),
                         'guest_session',
                         'new')
    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key)
    normalize_response(response,
                       date_time_format=date_time_format)
    return response


def base_method_url(api_base_url: str) -> str:
    return urljoin(api_base_url, 'authentication')


def normalize_response(response: Dict[str, Any],
                       *,
                       date_time_format: str) -> None:
    try:
        expiration_date_time_string = response['expires_at']
    except KeyError:
        return

    response['expires_at'] = datetime.strptime(expiration_date_time_string,
                                               date_time_format)



import pytest

from tests import strategies
from tests.utils import example


@pytest.fixture(scope='session')
def api_key() -> str:
    return example(strategies.api_keys)


@pytest.fixture(scope='session')
def invalid_api_key() -> str:
    return example(strategies.invalid_api_keys)


@pytest.fixture(scope='session')
def api_base_url() -> str:
    return example(strategies.api_base_urls)



from asyncio import (AbstractEventLoop,
                     ensure_future)
from datetime import (date,
                      datetime)
from typing import (Any,
                    Dict,
                    Set)

import pytest
from aiohttp import ClientSession

from asynctmdb.methods import movies
from tests import strategies
from tests.utils import example


@pytest.fixture(scope='function')
def movie_details(api_base_url: str,
                  api_key: str,
                  session: ClientSession,
                  event_loop: AbstractEventLoop) -> Dict[str, Any]:
    while True:
        movie_id = example(strategies.movies_ids)
        coroutine = movies.details(movie_id,
                                   api_base_url=api_base_url,
                                   api_key=api_key,
                                   session=session)
        future = ensure_future(coroutine)
        result = event_loop.run_until_complete(future)
        try:
            if not result['imdb_id']:
                continue
        except KeyError:
            continue

        return result


@pytest.fixture(scope='session')
def min_movie_id() -> int:
    return strategies.MIN_MOVIE_ID


@pytest.fixture(scope='function')
def movie_id(movie_details: Dict[str, Any]) -> int:
    return movie_details['id']


@pytest.fixture(scope='function')
def nonexistent_movie_id() -> int:
    return example(strategies.nonexistent_movies_ids)


@pytest.fixture(scope='session')
def min_movie_release_date() -> date:
    return date(year=1890, month=1, day=1)


@pytest.fixture(scope='session')
def movie_changes_start_date() -> datetime:
    return example(strategies.movies_changes_date_times)


@pytest.fixture(scope='session')
def movie_changes_end_date(movie_changes_start_date: datetime) -> datetime:
    return movie_changes_start_date + movies.MAX_TIME_RANGE_LENGTH


@pytest.fixture(scope='session')
def movie_videos_sites() -> Set[str]:
    return {'YouTube'}


@pytest.fixture(scope='function')
def page_number() -> int:
    return example(strategies.pages_numbers)


@pytest.fixture(scope='function')
def invalid_page_number() -> int:
    return example(strategies.invalid_pages_numbers)


@pytest.fixture(scope='function')
def movie_rating() -> float:
    return example(strategies.movies_ratings)



from contextlib import suppress
from datetime import datetime
from typing import (Any,
                    Dict)

from aiohttp import ClientSession

from asynctmdb.common import (DATE_FORMAT,
                              StatusCode)
from asynctmdb.config import API_BASE_URL
from asynctmdb.methods import find


async def movie(imdb_id: str,
                *,
                api_base_url: str = API_BASE_URL,
                api_key: str,
                language: str = None,
                session: ClientSession) -> Dict[str, Any]:
    params = {}
    if language is not None:
        params['language'] = language

    response = await find.by['imdb_id'](imdb_id,
                                        api_base_url=api_base_url,
                                        api_key=api_key,
                                        session=session,
                                        **params)
    try:
        records = response['movie_results']
    except KeyError as err:
        status_code = response['status_code']
        if status_code != StatusCode.RESOURCE_NOT_FOUND:
            return response
        err_msg = 'Invalid IMDb id: "{imdb_id}".'
        raise ValueError(err_msg) from err

    try:
        record, = records
    except ValueError as err:
        if records:
            err_msg = (f'Movie imdb id "{imdb_id}" is ambiguous: '
                       f'found {len(records)} records.')
        else:
            err_msg = ('No record found for movie with '
                       f'IMDb id "{imdb_id}".')
        raise ValueError(err_msg) from err
    else:
        normalize_movie(record)
        return record


def normalize_movie(record: Dict[str, Any],
                    *,
                    format_string: str = DATE_FORMAT) -> None:
    with suppress(TypeError):
        record['release_date'] = (datetime.strptime(record['release_date'],
                                                    format_string)
                                  .date())



from setuptools import (setup,
                        find_packages)

from asynctmdb.config import PROJECT_NAME

project_base_url = 'https://github.com/lycantropos/asynctmdb/'
setup(name=PROJECT_NAME,
      packages=find_packages(exclude=('tests',)),
      version='0.0.1',
      description='Asynchronous working with TMDB API.',
      author='Azat Ibrakov',
      author_email='azatibrakov@gmail.com',
      url=project_base_url,
      download_url=project_base_url + 'archive/master.tar.gz',
      install_requires=[
          'aiohttp>=2.2.5',  # asynchronous HTTP
      ],
      setup_requires=['pytest-runner>=2.11'],
      tests_require=[
          'pydevd>=1.0.0',  # debugging
          'pytest>=3.0.5',
          'pytest-asyncio>=0.6.0',  # "asyncio" support
          'pytest-cov>=2.4.0',  # coverage
          'hypothesis>=3.13.0',
      ])



import os
import string

from hypothesis import strategies

from asynctmdb.config import API_BASE_URL

api_key = os.environ['TMDb.API.Key']
api_base_urls = strategies.just(API_BASE_URL)
api_keys = strategies.just(api_key)
non_heximal_digits_characters = strategies.characters(
        blacklist_characters=string.hexdigits)
invalid_api_keys = strategies.text(alphabet=non_heximal_digits_characters)



import operator
from contextlib import suppress
from datetime import (datetime,
                      timedelta)
from itertools import chain
from typing import (Any,
                    Iterable,
                    Dict)

from aiohttp import ClientSession

from asynctmdb import requests
from asynctmdb.common import (DATE_FORMAT,
                              TIME_FORMAT,
                              DATE_TIME_FORMAT)
from asynctmdb.config import API_BASE_URL
from asynctmdb.utils import urljoin

MAX_TIME_RANGE_LENGTH = timedelta(days=14)
RELEASE_DATE_TIME_FORMAT = f'{DATE_FORMAT}T{TIME_FORMAT}.%fZ'
NOW_PLAYING_DATES_KEYS = ['minimum', 'maximum']


async def details(movie_id: int,
                  *,
                  api_base_url: str = API_BASE_URL,
                  api_key: str,
                  language: str = None,
                  session: ClientSession) -> Dict[str, Any]:
    method_url = base_method_url(api_base_url, movie_id)

    params = {}
    if language is not None:
        params['language'] = language

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    return response


async def account_states(movie_id: int,
                         *,
                         api_base_url: str = API_BASE_URL,
                         api_key: str,
                         session_id: str,
                         session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(base_method_url(api_base_url, movie_id),
                         'account_states')
    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  session_id=session_id)
    return response


async def alternative_titles(movie_id: int,
                             *,
                             api_base_url: str = API_BASE_URL,
                             api_key: str,
                             country: str = None,
                             session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(base_method_url(api_base_url, movie_id),
                         'alternative_titles')

    params = {}
    if country is not None:
        params['country'] = country

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    return response


async def changes(movie_id: int,
                  *,
                  api_base_url: str = API_BASE_URL,
                  api_key: str,
                  start_date: datetime = None,
                  end_date: datetime = None,
                  page: int = None,
                  session: ClientSession,
                  format_string: str = DATE_TIME_FORMAT) -> Dict[str, Any]:
    method_url = urljoin(base_method_url(api_base_url, movie_id),
                         'changes')

    if start_date is not None and end_date is not None:
        if end_date - start_date > MAX_TIME_RANGE_LENGTH:
            err_msg = ('Invalid date range: '
                       'should be a range no longer '
                       f'than {MAX_TIME_RANGE_LENGTH.days} days.')
            raise ValueError(err_msg)

    params = {}
    if start_date is not None:
        params['start_date'] = start_date.strftime(format_string)
    if end_date is not None:
        params['end_date'] = end_date.strftime(format_string)
    if page is not None:
        params['page'] = page

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    normalize_changes(response,
                      format_string=format_string)
    return response


async def credits(movie_id: int,
                  *,
                  api_base_url: str = API_BASE_URL,
                  api_key: str,
                  session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(base_method_url(api_base_url, movie_id),
                         'credits')
    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key)
    return response


async def images(movie_id: int,
                 *,
                 api_base_url: str = API_BASE_URL,
                 api_key: str,
                 language: str = None,
                 session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(base_method_url(api_base_url, movie_id),
                         'images')

    params = {}
    if language is not None:
        params['language'] = language

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    return response


async def keywords(movie_id: int,
                   *,
                   api_base_url: str = API_BASE_URL,
                   api_key: str,
                   session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(base_method_url(api_base_url, movie_id),
                         'keywords')
    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key)
    return response


async def release_dates(movie_id: int,
                        *,
                        api_base_url: str = API_BASE_URL,
                        api_key: str,
                        session: ClientSession,
                        format_string: str = RELEASE_DATE_TIME_FORMAT
                        ) -> Dict[str, Any]:
    method_url = urljoin(base_method_url(api_base_url, movie_id),
                         'release_dates')
    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key)
    normalize_release_dates(response,
                            format_string=format_string)
    return response


async def videos(movie_id: int,
                 *,
                 api_base_url: str = API_BASE_URL,
                 api_key: str,
                 language: str = None,
                 session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(base_method_url(api_base_url, movie_id),
                         'videos')

    params = {}
    if language is not None:
        params['language'] = language

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    return response


async def translations(movie_id: int,
                       *,
                       api_base_url: str = API_BASE_URL,
                       api_key: str,
                       session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(base_method_url(api_base_url, movie_id),
                         'translations')

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key)
    return response


async def recommendations(movie_id: int,
                          *,
                          api_base_url: str = API_BASE_URL,
                          api_key: str,
                          language: str = None,
                          page: int = None,
                          session: ClientSession,
                          format_string: str = DATE_FORMAT) -> Dict[str, Any]:
    method_url = urljoin(base_method_url(api_base_url, movie_id),
                         'recommendations')

    params = {}
    if language is not None:
        params['language'] = language
    if page is not None:
        params['page'] = page

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    normalize_recommendations(response,
                              format_string=format_string)
    return response


async def similar(movie_id: int,
                  *,
                  api_base_url: str = API_BASE_URL,
                  api_key: str,
                  language: str = None,
                  page: int = None,
                  session: ClientSession,
                  format_string: str = DATE_FORMAT) -> Dict[str, Any]:
    method_url = urljoin(base_method_url(api_base_url, movie_id),
                         'similar')

    params = {}
    if language is not None:
        params['language'] = language
    if page is not None:
        params['page'] = page

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    normalize_similar(response,
                      format_string=format_string)
    return response


async def reviews(movie_id: int,
                  *,
                  api_base_url: str = API_BASE_URL,
                  api_key: str,
                  language: str = None,
                  page: int = None,
                  session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(base_method_url(api_base_url, movie_id),
                         'reviews')

    params = {}
    if language is not None:
        params['language'] = language
    if page is not None:
        params['page'] = page

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    return response


async def lists(movie_id: int,
                *,
                api_base_url: str = API_BASE_URL,
                api_key: str,
                language: str = None,
                page: int = None,
                session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(base_method_url(api_base_url, movie_id),
                         'lists')

    params = {}
    if language is not None:
        params['language'] = language
    if page is not None:
        params['page'] = page

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    return response


async def rate(movie_id: int,
               *,
               rating: float,
               api_base_url: str = API_BASE_URL,
               api_key: str,
               guest_session_id: str = None,
               session_id: str = None,
               session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(base_method_url(api_base_url, movie_id),
                         'rating')

    params = {}
    if guest_session_id is not None:
        params['guest_session_id'] = guest_session_id
    if session_id is not None:
        params['session_id'] = session_id

    response = await requests.post(method_url=method_url,
                                   session=session,
                                   api_key=api_key,
                                   **params,
                                   json_body={'value': rating})
    return response


async def delete_rating(movie_id: int,
                        *,
                        api_base_url: str = API_BASE_URL,
                        api_key: str,
                        guest_session_id: str = None,
                        session_id: str = None,
                        session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(base_method_url(api_base_url, movie_id),
                         'rating')

    params = {}
    if guest_session_id is not None:
        params['guest_session_id'] = guest_session_id
    if session_id is not None:
        params['session_id'] = session_id

    response = await requests.delete(method_url=method_url,
                                     session=session,
                                     api_key=api_key,
                                     **params)
    return response


async def latest(*,
                 api_base_url: str = API_BASE_URL,
                 api_key: str,
                 language: str = None,
                 session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(api_base_url, 'movie', 'latest')

    params = {}
    if language is not None:
        params['language'] = language

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    return response


async def now_playing(*,
                      api_base_url: str = API_BASE_URL,
                      api_key: str,
                      language: str = None,
                      page: int = None,
                      region: str = None,
                      session: ClientSession,
                      format_string: str = DATE_FORMAT,
                      dates_keys: Iterable[str] = NOW_PLAYING_DATES_KEYS
                      ) -> Dict[str, Any]:
    method_url = urljoin(api_base_url, 'movie', 'now_playing')

    params = {}
    if language is not None:
        params['language'] = language
    if page is not None:
        params['page'] = page
    if region is not None:
        params['region'] = region

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    normalize_now_playing(response,
                          dates_keys=dates_keys,
                          format_string=format_string)
    return response


async def popular(*,
                  api_base_url: str = API_BASE_URL,
                  api_key: str,
                  language: str = None,
                  page: int = None,
                  region: str = None,
                  session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(api_base_url, 'movie', 'popular')

    params = {}
    if language is not None:
        params['language'] = language
    if page is not None:
        params['page'] = page
    if region is not None:
        params['region'] = region

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    return response


async def top_rated(*,
                    api_base_url: str = API_BASE_URL,
                    api_key: str,
                    language: str = None,
                    page: int = None,
                    region: str = None,
                    session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(api_base_url, 'movie', 'top_rated')

    params = {}
    if language is not None:
        params['language'] = language
    if page is not None:
        params['page'] = page
    if region is not None:
        params['region'] = region

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    return response


async def upcoming(*,
                   api_base_url: str = API_BASE_URL,
                   api_key: str,
                   language: str = None,
                   page: int = None,
                   region: str = None,
                   session: ClientSession) -> Dict[str, Any]:
    method_url = urljoin(api_base_url, 'movie', 'upcoming')

    params = {}
    if language is not None:
        params['language'] = language
    if page is not None:
        params['page'] = page
    if region is not None:
        params['region'] = region

    response = await requests.get(method_url=method_url,
                                  session=session,
                                  api_key=api_key,
                                  **params)
    return response


def base_method_url(api_base_url: str,
                    movie_id: int) -> str:
    return urljoin(api_base_url, 'movie', str(movie_id))


items_record = operator.itemgetter('items')


def normalize_changes(response: Dict[str, Any],
                      *,
                      format_string: str) -> None:
    try:
        results = response['changes']
    except KeyError:
        return
    records = chain.from_iterable(map(items_record, results))
    for record in records:
        record['time'] = datetime.strptime(record['time'],
                                           format_string)


release_dates_record = operator.itemgetter('release_dates')


def normalize_release_dates(response: Dict[str, Any],
                            *,
                            format_string: str) -> None:
    try:
        results = response['results']
    except KeyError:
        return

    records = chain.from_iterable(map(release_dates_record, results))
    for record in records:
        record['release_date'] = datetime.strptime(record['release_date'],
                                                   format_string)


def normalize_recommendations(response: Dict[str, Any],
                              *,
                              format_string: str) -> None:
    try:
        results = response['results']
    except KeyError:
        return

    for result in results:
        with suppress(TypeError):
            result['release_date'] = (datetime.strptime(result['release_date'],
                                                        format_string)
                                      .date())


normalize_similar = normalize_recommendations


def normalize_now_playing(response: Dict[str, Any],
                          *,
                          format_string: str,
                          dates_keys: Iterable[str]
                          ) -> None:
    try:
        dates = response['dates']
    except KeyError:
        return

    for key in dates_keys:
        dates[key] = (datetime.strptime(dates[key],
                                        format_string)
                      .date())



import string

from hypothesis import strategies

invalid_imdb_ids_characters = strategies.sampled_from(
        sorted(set(string.printable) - {'t'}))
invalid_imdb_ids = strategies.text(alphabet=invalid_imdb_ids_characters)



import abc
from typing import (Generic,
                    T_co)


def method_defined(cls: type,
                   *,
                   method_name: str) -> bool:
    return any(method_name in cls_or_ancestor.__dict__
               for cls_or_ancestor in cls.__mro__)


class AbstractAsyncContextManager(abc.ABC):
    """An abstract base class for async context managers."""

    def __aenter__(self):
        """Return `self` upon entering the runtime context."""
        return self

    @abc.abstractmethod
    def __aexit__(self, exc_type, exc_value, traceback):
        """Raise any exception triggered within the runtime context."""
        return None

    @classmethod
    def __subclasshook__(cls, subclass):
        if cls is AbstractAsyncContextManager:
            aenter_is_defined = method_defined(subclass,
                                               method_name='__aenter__')
            aexit_is_defined = method_defined(subclass,
                                              method_name='__aexit__')
            if aenter_is_defined and aexit_is_defined:
                return True
        return NotImplemented


class AsyncContextManager(Generic[T_co],
                          extra=AbstractAsyncContextManager):
    __slots__ = ()



import operator

import pytest
from aiohttp import ClientSession

from asynctmdb.methods import genres
from tests.utils import (is_positive_integer,
                         is_non_empty_string)


@pytest.mark.asyncio
async def test_movie_genres(api_base_url: str,
                            api_key: str,
                            session: ClientSession) -> None:
    records = await genres.movie(api_base_url=api_base_url,
                                 api_key=api_key,
                                 session=session)

    records_count = len(records)
    records_ids = list(map(operator.itemgetter('id'), records))
    records_names = list(map(operator.itemgetter('name'), records))

    assert isinstance(records, list)
    assert all(map(is_positive_integer, records_ids))
    assert all(map(is_non_empty_string, records_names))

    assert len(set(records_ids)) == records_count
    assert len(set(records_names)) == records_count


@pytest.mark.asyncio
async def test_tv_genres(api_base_url: str,
                         api_key: str,
                         session: ClientSession) -> None:
    records = await genres.tv(api_base_url=api_base_url,
                              api_key=api_key,
                              session=session)

    records_count = len(records)
    records_ids = list(map(operator.itemgetter('id'), records))
    records_names = list(map(operator.itemgetter('name'), records))

    assert isinstance(records, list)
    assert all(map(is_positive_integer, records_ids))
    assert all(map(is_non_empty_string, records_names))

    assert len(set(records_ids)) == records_count
    assert len(set(records_names)) == records_count



from datetime import (time,
                      datetime)
from typing import Iterator

from hypothesis import strategies

from tests.utils import example
from .api import (api_base_urls,
                  api_keys)
from .utils import (TMDB_FOUNDATION_DATE,
                    load_max_movie_id)

MIN_MOVIE_ID = 2
MAX_MOVIE_ID = load_max_movie_id(api_base_url=example(api_base_urls),
                                 api_key=example(api_keys))
MIN_PAGE_NUMBER = 1
MAX_PAGE_NUMBER = 1_000


def ratings(min_movie_rating: float = 0.5,
            max_movie_rating: float = 10.,
            movie_rating_step: float = 0.5) -> Iterator[float]:
    rating = min_movie_rating
    while rating <= max_movie_rating:
        yield rating
        rating += movie_rating_step


movies_ids = strategies.integers(min_value=MIN_MOVIE_ID,
                                 max_value=MAX_MOVIE_ID)
nonexistent_movies_ids = (strategies.integers(max_value=MIN_MOVIE_ID - 1)
                          # we're expecting
                          # that there will be less
                          # than 1_000 new records in TMDb
                          # while tests session is running
                          | strategies.integers(
        min_value=MAX_MOVIE_ID + 1_000))
movies_changes_date_times = strategies.datetimes(
        min_datetime=datetime.combine(TMDB_FOUNDATION_DATE,
                                      time()),
        max_datetime=datetime.utcnow())
movies_ratings = strategies.sampled_from(list(ratings()))
pages_numbers = strategies.integers(min_value=MIN_PAGE_NUMBER,
                                    max_value=MAX_PAGE_NUMBER)
invalid_pages_numbers = (strategies.integers(max_value=MIN_PAGE_NUMBER - 1)
                         | strategies.integers(min_value=MAX_PAGE_NUMBER + 1))
